@@ -10,7 +10,7 @@ module Faraday
       super(app)
       @logger = logger || begin
         require 'logger'
-        ::Logger.new(STDOUT)
+        ::Logger.new($stdout)
       end
       @filter = []
       @options = DEFAULT_OPTIONS.merge(options)
